""" Functions for reading in PEAKS search results.
"""
from copy import deepcopy

import pandas as pd
import polars as pl

from inspire.constants import (
    ACCESSION_KEY,
    CHARGE_KEY,
    DELTA_SCORE_KEY,
    ENGINE_SCORE_KEY,
    KNOWN_PTM_WEIGHTS,
    LABEL_KEY,
    MASS_DIFF_KEY,
    PEPTIDE_KEY,
    PTM_ID_KEY,
    PTM_IS_VAR_KEY,
    PTM_NAME_KEY,
    PTM_SEQ_KEY,
    PTM_WEIGHT_KEY,
    RT_KEY,
    SCAN_KEY,
    SEQ_LEN_KEY,
    SOURCE_KEY,
)
from inspire.utils import filter_for_prosit, remove_source_suffixes

# Define the relevant column names from PEAKS DB search results.
PEAKS_ACCESSION_KEY = 'Accession'
PEAKS_ASCORE_KEY = 'AScore'
PEAKS_CHARGE_KEY = 'Z'
PEAKS_CHIMERA_KEY = 'from Chimera'
PEAKS_LEN_KEY = 'Length'
PEAKS_MASS_KEY = 'Mass'
PEAKS_MZ_KEY = 'm/z'
PEAKS_PEPTIDE_KEY = 'Peptide'
PEAKS_PPM_KEY = 'ppm'
PEAKS_PTM_KEY = 'PTM'
PEAKS_RETENTION_TIME_KEY = 'RT'
PEAKS_SCAN_KEY = 'Scan'
PEAKS_SCORE_KEY = '-10lgP'
PEAKS_SOURCE_KEY = 'Source File'
PEAKS_RELEVANT_COLUMNS = [
    PEAKS_ACCESSION_KEY,
    PEAKS_ASCORE_KEY,
    PEAKS_CHARGE_KEY,
    PEAKS_CHIMERA_KEY,
    PEAKS_LEN_KEY,
    PEAKS_MASS_KEY,
    PEAKS_MZ_KEY,
    PEAKS_PEPTIDE_KEY,
    PEAKS_PPM_KEY,
    PEAKS_PTM_KEY,
    PEAKS_RETENTION_TIME_KEY,
    PEAKS_SCAN_KEY,
    PEAKS_SCORE_KEY,
    PEAKS_SOURCE_KEY,
]
PEAKS_11_RELEVANT_COLUMNS = [
    'Accession',
    'AScore',
    'z',
    # No chimera
    'Length',
    'Mass',
    'm/z',
    'Peptide',
    'ppm',
    'PTM',
    'RT',
    'Scan',
    '-10LgP',
    'Source File',
]

ID_NUMBERS = {
    'Deamidation (NQ)': 4,
    'Deamidation (N)': 4,
    'Deamidation (Q)': 4,
    'Phospho (S)': 5,
    'Phospho (T)': 5,
    'Phospho (Y)': 5,
    'Acetylation (N-term)': 3,
    'Acetylation (Protein N-term)': 3,
    'Oxidation (M)': 2,
    'Carbamidomethyl (C)': 1,
    'Carbamidomethylation': 1,
}

def _extract_terminus_ptms(mod_list, var_mods, terminus):
    """ Helper function to extract any ptms at the N or C termini.

    Parameters
    ----------
    mod_list : list of str
        A list of the ptms for the sequence.
    var_mods : dict
        A dictionary mapping ptm names to integer IDs.
    terminus : str
        A flag for the terminus (either N or C).

    Returns
    -------
    mod_list : list of str
        The updated list of the ptms for the sequence without the terminus modification.
    term_flag : str
        A ID of the terminus ptm.
    """
    term_mod = -1
    term_flag = '0'
    for mod in mod_list:
        if mod.endswith(f'({terminus}-term)'):
            term_mod = mod
            break
    if term_mod != -1:
        term_flag = f'{var_mods[term_mod]}'
        mod_list.remove(term_mod)
    return mod_list, term_flag

def separate_peaks_ptms(df_row, var_mods):
    """ Helper function to remove ptm markers from the Peaks Peptide column and create a
        separate ptm_seq column containing ptm data.

    Parameters
    ----------
    df_row : pd.Series
        A row of the Peaks search results DataFrame.
    var_mods : dict
        A dictionary mapping ptm names to integer ids.

    Returns
    -------
    df_row : pd.Series
        The updated row with peptide and ptms separated.
    """
    results = {}
    peptide = df_row[PEAKS_PEPTIDE_KEY]
    ptms = df_row[PEAKS_ASCORE_KEY]
    ptm_prefix = '0'
    ptm_suffix = '0'

    if isinstance(ptms, str) and ptms:
        ptm_list = [a.strip(' ') for a in ptms.split(';')]
        ptm_list = [b.split(':')[1].strip(' ') for b in ptm_list]
    else:
        results[PEPTIDE_KEY] = peptide
        results[PTM_SEQ_KEY] = None
        return results
    pep_seq = ''
    ptm_seq = ''
    idx = 0

    while peptide:
        assert peptide[0] != '('
        if len(peptide) == 1:
            pep_seq += peptide[0]
            ptm_seq += '0'
            break

        if peptide[1] != '(':
            pep_seq += peptide[0]
            ptm_seq += '0'
            peptide = peptide[1:]
        else:
            pep_seq += peptide[0]
            peptide = peptide[1:]

            end_mod = peptide.index(')') + 1
            mod_list = []
            while peptide[0] == '(':
                end_mod = peptide.index(')') + 1

                mod_list.append(ptm_list[0])
                ptm_list = ptm_list[1:]

                peptide = peptide[end_mod:]
                if not peptide:
                    break

            if idx == 0:
                mod_list, ptm_prefix = _extract_terminus_ptms(mod_list, var_mods, 'N')

            if not peptide:
                mod_list, ptm_suffix = _extract_terminus_ptms(mod_list, var_mods, 'C')

            if mod_list:
                ptm_seq += str(var_mods[mod_list[0]])
            else:
                ptm_seq += '0'
        idx += 1

    results[PEPTIDE_KEY] = pep_seq
    results[PTM_SEQ_KEY] = '.'.join([ptm_prefix, ptm_seq, ptm_suffix])
    return results

def collect_peaks_var_mods(peaks_df):
    """ Helper function to collect all of the ptms present in PEAKS DB search results.

    Parameters
    ----------
    peaks_df : pd.DataFrame
        A DataFrame of PEAKS DB search results.

    Returns
    -------
    ptms_df : pd.DataFrame
        A small DataFrame listing the unique ptms found in the data.
    """
    var_mod_df = deepcopy(peaks_df[[PEAKS_PTM_KEY, PEAKS_PEPTIDE_KEY]])
    var_mod_df = var_mod_df.with_columns(
        pl.col(PEAKS_PTM_KEY).apply(
            lambda x: None if not isinstance(x, str) else [a.strip(' ') for a in x.split(';')],
            skip_nulls=False,
        ).alias('ptm_names')
    )

    var_mod_df = var_mod_df.with_columns(
        pl.col(PEAKS_PEPTIDE_KEY).str.replace_all(r'[A-Za-z]', '').apply(
            lambda x : [a.strip('(').strip(')') for a in x.split(')(')]
        ).alias('peaks_ptm_weights')
    )

    var_mod_df = var_mod_df.filter(
        (pl.col('ptm_names').is_not_null()) & (pl.col('peaks_ptm_weights').is_not_null())
    )

    if not var_mod_df.shape[0]:
        return pd.DataFrame({
            PTM_NAME_KEY: [],
            PTM_WEIGHT_KEY: [],
            PTM_ID_KEY: [],
        })

    var_mod_df = var_mod_df.with_columns(
        pl.struct(['ptm_names', 'peaks_ptm_weights']).apply(
            lambda x : ["~".join(y) for y in zip(x['ptm_names'], x['peaks_ptm_weights'])]
        ).alias('combined_ptm_data')
    )
    ptms = [x.split('~') for x in var_mod_df['combined_ptm_data'].explode().unique()]
    ptms = [x for x in ptms if x[0]]

    ptm_names = [name for name, _ in ptms]

    ptm_weights = [KNOWN_PTM_WEIGHTS.get(name, float(weight)) for name, weight in ptms]

    ptms_df = pd.DataFrame({
        PTM_NAME_KEY: pd.Series(ptm_names),
        PTM_WEIGHT_KEY: pd.Series(ptm_weights),
    })
    ptms_df = ptms_df.drop_duplicates(PTM_NAME_KEY).reset_index(drop=True)

    ptms_df[PTM_ID_KEY] = ptms_df[PTM_NAME_KEY].apply(
        lambda x : ID_NUMBERS.get(x, 9)
    )
    ptms_df = ptms_df.sort_values(by=PTM_ID_KEY).reset_index(drop=True)
    ptms_df[PTM_IS_VAR_KEY] = True

    return ptms_df

def read_single_peaks_data(df_loc):
    """ Function to read in PEAKS DB search results from a single file.

    Parameters
    ----------
    df_loc : str
        A location of PEAKS DB search results.

    Returns
    -------
    hits_df : pd.DataFrame
        A DataFrame of all search results properly formatted for inSPIRE.
    mods_dfs : pd.DataFrame
        A small DataFrame detailing the ptms found in the data.
    """
    try:
        peaks_df = pl.read_csv(df_loc, columns=PEAKS_RELEVANT_COLUMNS)
    except pl.exceptions.ColumnNotFoundError:
        peaks_df = pl.read_csv(df_loc, columns=PEAKS_11_RELEVANT_COLUMNS)
        peaks_df = peaks_df.with_columns(pl.lit('No').alias('from Chimera'))
        peaks_df = peaks_df.rename({'z': 'Z', '-10LgP': '-10lgP'})


    # Separate PTMs.
    var_mods = collect_peaks_var_mods(peaks_df)
    var_mod_dict = dict(zip(var_mods[PTM_NAME_KEY].tolist(), var_mods[PTM_ID_KEY].tolist()))

    if var_mod_dict:
        peaks_df = peaks_df.with_columns(
            pl.struct([PEAKS_PEPTIDE_KEY, PEAKS_ASCORE_KEY]).apply(
                lambda x: separate_peaks_ptms(x, var_mod_dict)
            ).alias('results')
        ).unnest('results')
        peaks_df = peaks_df.drop(PEAKS_PEPTIDE_KEY)
    else:
        peaks_df = peaks_df.rename(
            {PEAKS_PEPTIDE_KEY: PEPTIDE_KEY}
        )
        peaks_df = peaks_df.with_columns(
            pl.lit(None).alias(PTM_SEQ_KEY)
        )

    print(peaks_df.columns)

    # Rename to match inSPIRE naming scheme.
    peaks_df = peaks_df.rename({
        PEAKS_ACCESSION_KEY: ACCESSION_KEY,
        PEAKS_CHARGE_KEY: CHARGE_KEY,
        PEAKS_LEN_KEY: SEQ_LEN_KEY,
        PEAKS_RETENTION_TIME_KEY: RT_KEY,
        PEAKS_SCORE_KEY: ENGINE_SCORE_KEY,
        PEAKS_PPM_KEY: MASS_DIFF_KEY,
    })

    # Filter for Prosit and add feature columns not present.
    peaks_df = peaks_df.with_columns(
        pl.col(PEAKS_CHIMERA_KEY).apply(
            lambda x : 1 if x == 'Yes' else 0
        ).alias('fromChimera'),
        pl.col(ACCESSION_KEY).fill_null(
            pl.lit('unknown'),
        ).alias(ACCESSION_KEY),
    )

    peaks_df = filter_for_prosit(peaks_df)
    peaks_df = peaks_df.with_columns(
        (pl.col(PEAKS_MASS_KEY)/pl.col(SEQ_LEN_KEY)).alias('avgResidueMass'),
        pl.lit(0).alias(DELTA_SCORE_KEY),
        pl.lit(0).alias('missedCleavages'),
        pl.col(PEAKS_SOURCE_KEY).apply(remove_source_suffixes).alias(SOURCE_KEY),
        pl.col(PEAKS_SCAN_KEY).apply(
            lambda x : x if isinstance(x, int) else int(x.split(':')[-1])
        ).alias(SCAN_KEY),
        pl.col(ACCESSION_KEY).apply(
            lambda x : -1 if isinstance(x, str) and ('DECOY' in x or 'rev' in x) else 1
        ).alias(LABEL_KEY),
    )

    for col, dtype in peaks_df.schema.items():
        if dtype == pl.Null:
            peaks_df = peaks_df.with_columns(
                pl.lit('').alias(col)
            )

    peaks_df = peaks_df.drop(
        [
            PEAKS_ASCORE_KEY,
            PEAKS_CHIMERA_KEY,
            PEAKS_MASS_KEY,
            PEAKS_MZ_KEY,
            PEAKS_PTM_KEY,
            PEAKS_SCAN_KEY,
            PEAKS_SOURCE_KEY,
        ]
    )

    return peaks_df, var_mods


def read_peaks_data(peaks_data):
    """ Function to read in PEAKS DB search results from one or more files.

    Parameters
    ----------
    peaks_data : str or list of str
        A single location of PEAKS DB search results or a list of locations.

    Returns
    -------
    hits_df : pd.DataFrame
        A DataFrame of all search results properly formatted for inSPIRE.
    mods_dfs : pd.DataFrame
        A small DataFrame detailing the ptms found in the data.
    """
    if isinstance(peaks_data, list):
        peaks_dfs = []
        variable_mods_dfs = []
        for peaks_file in peaks_data:
            hits_df, mods_df = read_single_peaks_data(peaks_file)
            peaks_dfs.append(hits_df)
            variable_mods_dfs.append(mods_df)

        peaks_dfs = [x.select(peaks_dfs[0].columns) for x in peaks_dfs]

        # Combine DataFrames and validate that same PTMs are present.
        hits_df = pl.concat(peaks_dfs)

        return hits_df, variable_mods_dfs[0]

    return read_single_peaks_data(peaks_data)



""" Functions to create figures needed for the inSPIRE report.
"""
import numpy as np
from plotly.colors import n_colors
import plotly.graph_objects as go
from plotly.subplots import make_subplots


def create_binders_fig(fdrs_df, search_engine):
    """ Function to create a plotly figure of percentage binders as predicted
        by NetMHCpan against q-value cut off.

    Parameters
    ----------
    fdrs_df : pd.DataFrame
        A DataFrame of counts of binders and PSMs at different q-values.
    search_engine : str
        The name of the original ms search engine used.

    Returns
    -------
    fig : str
        A plot of the percentage predicted binders against q-value converted to html.
    """
    trace1 = go.Scatter(
        x=fdrs_df['FDR'],
        y=fdrs_df['SpirePercentageBinders'],
        name='inSPIRE Percentage Binders',
        line={'color': 'red'},
        mode='lines+markers',
        connectgaps=True,
    )
    trace2 = go.Scatter(
        x=fdrs_df['FDR'],
        y=fdrs_df['searchEnginePercentageBinders'],
        name='Original Percentage Binders',
        line={'color': 'black'},
        mode='lines+markers',
        connectgaps=True,
    )

    data = [trace1, trace2]
    layout2 = go.Layout(
        yaxis={'title': 'Percentage of all Identified PSMs'},
        xaxis={'title': 'q-value'},
    )
    fig = go.Figure(data=data, layout=layout2)
    fig.update_layout(
        plot_bgcolor='rgba(0,0,0,0)',
        title=f'Percentage Binders Identified with {search_engine.title()} and inSPIRE',
        width=1000,
        height=500,
        title_x=0.5
    )
    fig.update_xaxes(showline=True, linewidth=0.5, linecolor='black')
    fig.update_yaxes(showline=True, linewidth=0.5, linecolor='black')

    return fig.to_html()


def create_psms_fig(fdrs_df, output_folder):
    """ Function for plotting the number of PSMs identified at varying q-values
        for inspire and the original search engine.

    Parameters
    ----------
    fdrs_df : pd.DataFrame
        A DataFrame of statistics for identifications at varying q-values.
    output_folder : str
        The folder where the PSM-fdr curve will be written to.

    Returns
    -------
    fig : str
        The plot of PSMs vs. q-value converted to html.
    """
    spire_trace = go.Scatter(
        x=fdrs_df['FDR'],
        y=fdrs_df['nSpirePsms'],
        name='With inSPIRE',
        line={'color': 'red'},
        mode='lines+markers',
        connectgaps=True,
    )
    search_engine_trace = go.Scatter(
        x=fdrs_df['FDR'],
        y=fdrs_df['nSearchEnginePsms'],
        name='Without inSPIRE',
        line={'color': 'black'},
        mode='lines+markers',
        connectgaps=True,
    )

    data = [spire_trace, search_engine_trace]
    layout = go.Layout(
        yaxis={'title':'Number of Identifications'},
        xaxis={'title': 'q-value'},
    )
    fig = go.Figure(data=data, layout=layout)


    fig.update_layout(
        plot_bgcolor='rgba(0,0,0,0)',
        title='PSMs Identified with and without inSPIRE',
        title_x=0.5,
    )
    fig.update_xaxes(showline=True, linewidth=0.5, linecolor='black')
    fig.update_yaxes(showline=True, linewidth=0.5, linecolor='black')
    fig.update_layout(
        width=500,
        height=300,
    )

    fig.write_image(
        f'{output_folder}/img/psm_fdr_curve.svg', engine='kaleido'
    )
    fig.update_layout(
        width=1000,
        height=500,
    )

    return fig.to_html()


def create_violin_fig(combined_df, most_positive_features, most_negative_features):
    """ Function to create a plotly figure of violin plots of the distributions of
        key features above and below the 1% FDR cut off.

    Parameters
    ----------
    combined_df : pd.DataFrame
        A DataFrame of inSPIRE output with input features.
    most_positive_features : list of str
        The most positively weighted features by percolator/mokapot.
    most_negative_features : list of str
        The most negatively weighted features by percolator/mokapot.

    Returns
    -------
    fig : str
        Plots of the distribution of each feature for accepted and rejected PSMs
        converted to html.
    """
    fig = make_subplots(
        rows=2,
        cols=3,
        subplot_titles=tuple(most_positive_features + most_negative_features),
        vertical_spacing=0.12,
    )

    for row_idx, feature_group in enumerate([most_positive_features, most_negative_features]):
        for idx, feature in enumerate(feature_group):
            fig.add_trace(
                go.Violin(
                    x=combined_df[combined_df['Status'] == 'Accepted']['Status'],
                    y=combined_df[combined_df['Status'] == 'Accepted'][feature],
                    # color=combined_df['Status'],
                    name='Accepted',
                    fillcolor='#8FBC8F',
                    line_color='black',
                    line_width=0.5,
                    marker_size=0.25,
                    scalegroup=(row_idx*3) + idx,
                    # box_visible=True,
                    meanline_visible=True,
                ),
                row=row_idx+1,
                col=idx+1,
            )
            fig.add_trace(
                go.Violin(
                    x=combined_df[combined_df['Status'] == 'Rejected']['Status'],
                    y=combined_df[combined_df['Status'] == 'Rejected'][feature],
                    fillcolor='#a52a2a',
                    line_color='black',
                    line_width=0.5,
                    marker_size=0.25,
                    # color=combined_df['Status'],
                    name='Rejected',
                    scalegroup=(row_idx*3) +idx,
                    # box_visible=True,
                    meanline_visible=True,
                ),
                row=row_idx+1,
                col=idx+1,
            )

    fig.update_layout(
        title='Distribution of High Importance Features',
        width=1100,
        height=700,
        title_x=0.5,
        showlegend=False,
        plot_bgcolor='rgba(0,0,0,0)',
    )
    fig.update_xaxes(showline=True, linewidth=0.5, linecolor='black')
    fig.update_yaxes(showline=True, linewidth=0.5, linecolor='black')

    return fig.to_html()

def _get_color_map(feature_weights, wt_max, wt_min):
    """ Helper function to get the color weighting of different features
        based on percolator feature importance.

    Parameters
    ----------
    feature_weights : pd.Series
        A Series of feature weights for a single fold.
    wt_max : float
        The maximum features weight across all folds.
    wt_min : float
        The minimum features weight across all folds.

    Returns
    -------
    col_weights : list of int
        A list of the value on the color spectrum for each feature.
    """
    col_weights = []
    for entry in feature_weights:
        if entry >= 0:
            normed_val = entry/wt_max
            col_weights.append(128 + int(round(127*normed_val)))
        else:
            normed_val = 1 - (entry/wt_min)
            col_weights.append(int(round(127*normed_val)))

    return col_weights

def create_weights_fig(
        weights_df
    ):
    """ Function to create a table of percolator/mokapot features and their
        weights.

    Parameters
    ----------
    weights_df : pd.DataFrame
        A Series of feature weights for a single fold.

    Returns
    -------
    fig : str
        A table of the features and their weights in Percolator converted to html.
    """
    colors1 = n_colors('rgb(255, 204, 204)', 'rgb(255, 255, 255)', 128, colortype='rgb')
    colors2 = n_colors('rgb(255, 255, 255)', 'rgb(204, 255, 204)', 128, colortype='rgb')
    colors = colors1 + colors2
    fig = go.Figure(
        data=[
            go.Table(
                header={
                    'values': [
                        'feature',
                        'weightFold1',
                        'weightFold2',
                        'weightFold3',
                        'averageWeight',
                    ],
                    'line_color': 'black',
                },
                cells={
                    'values': [
                        weights_df.feature,
                        weights_df.weightFold1,
                        weights_df.weightFold2,
                        weights_df.weightFold3,
                        [round(x, 3) for x in weights_df.averageWeight],
                    ],
                    'fill_color': [
                        np.array(['#FFFFFF']*len(weights_df)),
                        np.array(colors)[_get_color_map(
                            weights_df.weightFold1,
                            weights_df.weightFold1.max(),
                            weights_df.weightFold1.min(),
                        )],
                        np.array(colors)[_get_color_map(
                            weights_df.weightFold2,
                            weights_df.weightFold2.max(),
                            weights_df.weightFold2.min(),
                        )],
                        np.array(colors)[_get_color_map(
                            weights_df.weightFold3,
                            weights_df.weightFold3.max(),
                            weights_df.weightFold3.min(),
                        )],
                        np.array(colors)[_get_color_map(
                            weights_df.averageWeight,
                            weights_df.averageWeight.max(),
                            weights_df.averageWeight.min(),
                        )],
                    ],
                    'line_color': 'black',
                }
            )
        ]
    )
    fig.update_layout(
        title='Weights of Features Selected by inSPIRE.',
        width=1000,
        height=1100,
        title_x=0.5
    )
    return fig.to_html()



""" Functions for generating the html report at the end of the PEPSeek
    pipeline.
"""
import os
import webbrowser

import pandas as pd

from inspire.constants import ENDC_TEXT, OKCYAN_TEXT

def create_epitope_report(config):
    """ Function to create the final html report and open it in the brower.

    Parameters
    ----------
    config : inspire.config.Config
        The Config object for the whole pipeline.
    figures : dict
        A dictionary containing all of the required plots.
    """
    out_path = os.path.abspath(config.output_folder)

    base_table = """<table style="width:40%">
        <tr>
            <th>Peptide</th>
            <th>Proteins</th>
        </tr>
    """
    ep_df = pd.read_csv(
        f'{config.output_folder}/PEPSeek/potentialEpitopeCandidates.csv'
    )

    for _, df_row in ep_df.iterrows():
        base_table += f'''
            <tr>
                <td>{df_row['peptide']}</td>
                <td>{df_row['protein']}</td>
            </tr>
        '''

    with open(f'{out_path}/img/PEPSeek_bar_plot.svg', mode='r', encoding='UTF-8') as in_f:
        bar_plot = in_f.read().strip('\n')

    with open(f'{out_path}/img/PEPSeek_metrics.svg', mode='r', encoding='UTF-8') as in_f:
        swarm_plot = in_f.read()

    base_table += '</table>'

    html_string = ('''
    <html>
        <head>
            <link 
                rel="stylesheet"
                href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.1/css/bootstrap.min.css"
            >
            <style>
                body{
                    font-family: Helvetica;
                    margin:0 100;
                    padding-bottom: 50px;
                    background:whitesmoke;
                }
                h2{
                    color: firebrick;
                    font-family: Helvetica;
                }
                h3{
                    color: firebrick;
                    font-family: Helvetica;
                }
                h2:after
                {
                    content:' ';
                    display: block;
                    border:0.25px solid #696969;
                    position: absolute;
                    width: 60%;
                    margin-top: 2px;
                    left: 20%;
                }
                table {
                    font-family: Helvetica;
                    width: 60%;
                    border: 2px solid #696969;
                }
                th, td {
                    border: 1px solid #696969;
                    padding: 2px;
                }
            </style>
        </head>
        <body>
            <center>
            <h2>PEPSeek Report for ''' + config.experiment_title + '''</h2>
            </center>
            <h3>
                Epitope Candidates Found
            </h3>
            <p>
                Below is a full table of the pathogen peptides identified and their associated
                proteins.
            </p>
            <p>
                ''' + base_table + '''
            </p>
            <h3>
                Shared and PEPSeek Only Peptides:
            </h3>
            <p>
                This figure compares the number of peptides identified by both the
                search engine and PEPSeek against peptides identified by PEPSeek alone.
            </p>
            <br>
             ''' + bar_plot +
    '''
            <h3>
                MS2 Quality Metrics
            </h3>
            <p>
                This figure shows distribution of spectral angle, engine score, and
                retention time prediction error on the pathogen peptide identifications.
                For spectral angle and retention time prediction error we should see a
                similar distribution. On engine score, the PEPSeek only peptides likely
                score lower.
            </p>
            <br>
    ''' + swarm_plot
    )
    if os.path.exists(f'{out_path}/img/logo_comp_plots.svg'):
        with open(f'{out_path}/img/logo_comp_plots.svg', mode='r', encoding='UTF-8') as in_f:
            logo_plot = in_f.read().strip('\n')
        html_string += (
            '''
                    <h3>
                        Pathogen vs. Host Amino Acid Frequency:
                    </h3>
                    <p>
                        This chart shows the JS divergence between amino acid frequencies in host
                        and pathogen peptides. Amino acids on the positive y-axis are
                        overrepresented in the pathogen peptides compared to the host,
                        while amino acids on the negative y-axis
                        are overrepresented in the host peptides compared to the pathogen.
                        This plot can provide insight into the differing characteristics
                        of pathogen and host peptides.
                    </p>
            ''' + logo_plot
        )
    if os.path.exists(f'{out_path}/img/PEPSeek_pca.svg'):
        with open(f'{out_path}/img/PEPSeek_pca.svg', mode='r', encoding='UTF-8') as in_f:
            ep_pca = in_f.read().strip('\n')
        html_string += (
            '''
                    <h3>
                        Quantitative Data:
                    </h3>
                    <p>
                        Dimensionality reduction applied to quantiative data from the pathogen
                        peptide and a random seleciton of host peptides below. Ideally we should
                        see some clustering effects with pathogen peptides mostly close together.
                        This is likely not a perfect clustering as there is noise in the
                        label free quantification, however it can be a useful indicator.
                    </p>
            ''' + ep_pca
        )

    if os.path.exists(f'{out_path}/img/PEPSeek_affinity_cluster.svg'):
        with open(
            f'{out_path}/img/PEPSeek_affinity_cluster.svg',
            mode='r',
            encoding='UTF-8',
        ) as in_f:
            aff_clust = in_f.read().strip('\n')
        html_string += (
            '''
                <h3>
                    Binding Affinity Predictions:
                </h3>
                <p>
                    The predicted binding affinities for the identified peptides. This can
                    help see you the distribution of predicted binders for each allele in
                    your host cell.
                </p>
            ''' + aff_clust
        )

    html_string += (
    '''
        <h3>
            MS2 Spectral Plots:
        </h3>
        <p>
            These are the MS2 spectra based on which the peptides were assigned. Inspection
            of the spectra can be informative and increase your confidence in peptides
            identified.
        </p>
        <embed src="
    ''' + f'{config.output_folder}/PEPSeek/spectralPlots.pdf" width=1000 height=2000>'
    )


    output_path = f'{config.output_folder}/PEPSeek/pepseek-report.html'
    with open(output_path, 'w', encoding='UTF-8') as output_file:
        output_file.write(html_string)

    print(
        OKCYAN_TEXT +
        '\tPEPSeek report generated.' +
        ENDC_TEXT
    )

    if not config.silent_execution:
        webbrowser.open(
            'file://' + os.path.realpath(output_path),
            new=2
        )



""" Functions from proteome remapping in relation to epitope candidate extraction.
"""
import multiprocessing as mp
import pandas as pd

from inspire.constants import (
    ENDC_TEXT,
    OKCYAN_TEXT,
)
from inspire.utils import fetch_proteome


def filter_pathogen_only_peptides(final_df, config):
    """ Function to remap and filter the final assignments for pathogen only peptides.

    Parameters
    ----------
    final_df : pd.DataFrame
        PEP filtered DataFrame.
    config : inspire.config.Config
        The Config object which controls the experiment.

    Returns
    -------
    final_df : pd.DataFrame
        DataFrame of pathogen mapped peptides only.
    multi_mapped_df : pd.DataFrame
        DataFrame of pathogen mapped peptides that multimapped.
    host_df : pd.DataFrame
        DataFrame of host peptides identified.
    """
    final_df = add_accession_data(final_df, config)

    total_count = final_df.shape[0]
    final_df, multi_mapped_df, host_df = filter_multi_mapped(final_df, config.host_proteome)
    no_host_accession_count = final_df.shape[0]

    print(
        OKCYAN_TEXT +
        f'\t{total_count - no_host_accession_count} dropped due to host accessions.'
        + ENDC_TEXT
    )

    if not no_host_accession_count:
        exit_with_no_peptides_identified(config)

    total_count = final_df.shape[0]
    final_df = final_df[final_df['pathogenProteins'].apply(lambda x : x != 'unknown')]
    known_accession_count = final_df.shape[0]

    print(
        OKCYAN_TEXT +
        f'\t{total_count - known_accession_count} PSMs dropped due to unknown accessions.'
        + ENDC_TEXT
    )

    if not known_accession_count:
        exit_with_no_peptides_identified(config)

    return final_df, multi_mapped_df, host_df


def _sub_add_accession(unique_df, host_proteome, pathogen_proteome):
    unique_df['hostAccessions'] = unique_df['peptide'].apply(
        lambda x : fetch_antigen_data(
            x, host_proteome, trace_accession=False,
        )
    )

    unique_df['pathogenAccessions'] = unique_df['peptide'].apply(
        lambda x : fetch_antigen_data(x, pathogen_proteome)
    )
    unique_df['pathogenProteins'] = unique_df['pathogenAccessions'].apply(
        lambda x : ','.join([y[0] for y in x]) if x else 'unknown'
    )
    return unique_df

def add_accession_data(final_df, config):
    """ Function to add accession data to the final assignments, remapping all peptides to
        the host and pathogen proteomes.

    Parameters
    ----------
    final_df : pd.DataFrame
        The inSPIRE final assignments DataFrame.
    config : inspire.config.Config
        The Config object for running the experiment.

    Returns
    -------
    final_df : pd.DataFrame
        The inSPIRE final assignments DataFrame updated with
    """
    host_proteome = fetch_proteome(config.host_proteome)
    pathogen_proteome = fetch_proteome(config.pathogen_proteome)

    unique_df = final_df[['peptide']].drop_duplicates(subset=['peptide'])

    batch_size = unique_df.shape[0]//config.n_cores

    batch_values = []
    for batch_idx in range(config.n_cores):
        batch_values.extend([batch_idx]*batch_size)

    if (additional_psms := unique_df.shape[0]%config.n_cores):
        batch_values.extend([config.n_cores-1]*additional_psms)

    unique_df['batch'] = batch_values

    pep_df_list = [unique_df[unique_df['batch'] == b_idx] for b_idx in range(config.n_cores)]
    func_args = []
    for sub_pep_df in pep_df_list:
        func_args.append([sub_pep_df, host_proteome, pathogen_proteome])

    with mp.get_context('spawn').Pool(processes=config.n_cores) as pool:
        pep_df_list = pool.starmap(_sub_add_accession, func_args)
    remapped_pep_df = pd.concat(pep_df_list)

    total_df = pd.merge(final_df, remapped_pep_df, how='inner', on='peptide')

    return total_df


def fetch_antigen_data(
        peptide,
        proteome,
        trace_accession=True,
    ):
    """ Function to check for the presence of an identified peptide as either canonical
        or spliced in the input proteome.
    """
    accession_stratum = []
    peptide = peptide.replace('I', 'L')
    for protein in proteome:
        if peptide in protein[1]:
            location = protein[1].index(peptide) + 1
            if trace_accession:
                if not accession_stratum:
                    accession_stratum = [(protein[0], location, protein[2])]
                else:
                    accession_stratum.append((protein[0], location, protein[2]))
            else:
                return 'PCP'
    if trace_accession:
        return accession_stratum
    return 'unknown'


def filter_multi_mapped(final_df, host_proteome_filename):
    """ Function to remove peptides which map to both host and pathogen proteomes.

    Parameters
    ----------
    final_df : pd.DataFrame
        A DataFrame of all peptides identified.
    host_proteome_filename : str
        The path to the file containing the host proteome.

    Returns
    -------
    final_df : pd.DataFrame
        A DataFrame of all peptides identified filterd to remove those from the
        host proteome.
    multi_mapped_df : pd.DataFrame
        A DataFrame of all peptides which map to both the host and pathogen proteomes.
    """
    host_proteome = fetch_proteome(host_proteome_filename)
    multi_mapped_df = final_df[
        (final_df['hostAccessions'] != 'unknown') &
        (final_df['pathogenProteins'] != 'unknown')
    ]
    multi_mapped_df['hostAccessions'] = multi_mapped_df['peptide'].apply(
        lambda x : fetch_antigen_data(
            x, host_proteome, trace_accession=False,
        )
    )
    host_df = final_df[
        (final_df['hostAccessions'] != 'unknown') &
        (final_df['pathogenProteins'] == 'unknown')
    ]
    final_df = final_df[
        final_df['hostAccessions'].apply(lambda x : x == 'unknown')
    ]

    return final_df, multi_mapped_df, host_df


def extract_protein_ids(pathogen_accessions):
    """ Function to extract protein ids.
    """
    accessions = pathogen_accessions.split(')')
    proteins = []

    for acc in accessions:
        if acc:
            proteins.append('_'.join(acc.strip('[').strip(']').split('_')[:-1]))

    return ','.join(proteins)


def exit_with_no_peptides_identified(config):
    """ Function to exit if no pathogen peptides are identified.

    Parameters
    ----------
    config : inspire.config.Config
        The Config object which controls the experiment.
    """
    print(
        OKCYAN_TEXT +
        '\tNo pathogen peptides identified.'
        + ENDC_TEXT
    )

    _ = pd.ExcelWriter( # pylint: disable=abstract-class-instantiated
        f'{config.output_folder}/PEPSeek/potentialEpitopeCandidates.xlsx',
        engine='xlsxwriter',
    )



""" Function to normalise quantifications by Skyline across raw files.

    Intermeidate columns created during execution of this script:
        - source_valid : 1/0 flag indicating if there is a valid measurement in this source file.
        - source_clean : raw values with null replacement if the measurement is invalid.
        - source_norm : the final normalised value for that source file.
"""
from math import log2
import numpy as np
import pandas as pd

from inspire.constants import PEPTIDE_KEY, ACCESSION_KEY

from inspire.quant.utils import plot_correlations, plot_quant_clustermap, plot_distros

AVERAGE_QUANT_KEY = 'average'
N_VALID_QUANT_KEY = 'nValid'

def normalise_intensities(config):
    """ Function to normalise intensities across RAW files with robust regression.

    Parameters
    ----------
    config : inspire.config.Config
        Config object for the whole experiment.
    """
    # Read in quantification data
    quant_df = pd.read_csv(
        f'{config.output_folder}/quant/quantified_per_file.csv',
    )
    meta_df = pd.read_csv(
        f'{config.output_folder}/quant/metadata.csv'
    )
    sources = meta_df['renamed'].tolist()

    # Find the valid quantifications.
    quant_df = count_valid_and_identified_quantifications(quant_df, sources, config)

    # Fill invalid quantifications with null.
    quant_df = clean_quantification_values(quant_df, sources, config)
    quant_df[
        [PEPTIDE_KEY, ACCESSION_KEY] + [f'{source}_clean' for source in sources]
    ].to_csv(f'{config.output_folder}/quant/cleaned_quantification.csv', index=False)

    # Select relevant columns.
    quant_df = quant_df[
        [PEPTIDE_KEY, ACCESSION_KEY, N_VALID_QUANT_KEY]
        + [f'{source}_clean' for source in sources]
        + [f'{source}_valid' for source in sources]
    ]

    for source in sources:
        quant_df[f'{source}_clean'] = quant_df[f'{source}_clean'].apply(
            lambda intensity : log2(intensity) if (
                intensity is not None and intensity > 0
            ) else None
        )

    # Calculate average.
    quant_df[AVERAGE_QUANT_KEY] = quant_df[
        [f'{source}_clean' for source in sources]
    ].mean(axis=1, skipna=True, numeric_only=True)

    global_median = quant_df[[f'{source}_clean' for source in sources]].median(axis=None)
    for source in sources:
        quant_df = equalize_medians(quant_df, source, global_median)

    plot_distros(quant_df, config, sources)

    # Write output to csv.
    quant_df[
        [PEPTIDE_KEY, ACCESSION_KEY] + [f'{source}_norm' for source in sources]
    ].to_csv(f'{config.output_folder}/quant/normalised_quantification.csv', index=False)

    # Plot correlations between normalised intensities.
    try:
        plot_correlations('normalised_quantification', 'norm', config)
    except Exception as e:
        print(f'Plotting noremd correlations failed with error code {e}')

    try:
        plot_quant_clustermap(config)
    except Exception as e:
        print(f'Plotting cluster map failed with error code {e}')


def count_valid_and_identified_quantifications(quant_df, sources, config):
    """ Function find the number of valid quantifications where peptide also
        identified in that file.

    Parameters
    ----------
    quant_df : pd.DataFrame
        DataFrame of quantified peptides.
    sources : list of str
        List of the raw files where the peptide was quantified.
    config : inspire.config.Config
        Config object for the whole experiment.

    Returns
    -------
    quant_df : pd.DataFrame
        DataFrame of quantifications with flags indicating if they have valid quantification
        in each file.
    """
    for source in sources:
        quant_df[f'{source}_valid'] = quant_df.apply(
            lambda df_row, src=source : (
                (df_row[f'{src}_idp'] > config.skyline_idp_cut_off) &
                (df_row[f'{src}_ratio'] > config.skyline_bg_ratio_cut_off)
                # & (df_row[f'{src}_identified'] == 1)
            ),
            axis=1,
        )
    quant_df[N_VALID_QUANT_KEY] = quant_df[[f'{source}_valid' for source in sources]].sum(axis=1)

    return quant_df

def clean_quantification_values(quant_df, sources, config):
    """ Function to set invalid quantifications to null.

    Parameters
    ----------
    quant_df : pd.DataFrame
        DataFrame of quantified peptides.
    sources : list of str
        List of the raw files where the peptide was quantified.
    config : inspire.config.Config
        Config object for the whole experiment.

    Returns
    -------
    quant_df : pd.DataFrame
        DataFrame of quantifications with quantifications including null replacement for
        invalid quantifications.
    """
    for source in sources:
        quant_df[f'{source}_clean'] = quant_df.apply(
            lambda df_row, src=source : df_row[f'{src}_raw'] if (
                (df_row[f'{src}_idp'] > config.skyline_idp_cut_off) &
                (df_row[f'{src}_ratio'] > config.skyline_bg_ratio_cut_off)
                # & (df_row[f'{src}_identified'] == 1)
            ) else None,
            axis=1,
        )

    return quant_df


def equalize_medians(quant_df, source, global_median):
    """ Function to apply equalize medians for raw file normalisation of intensites
        from a given raw file.

    Parameters
    ----------
    quant_df : pd.DataFrame
        DataFrame of quantified peptides.
    source : list of str
        The raw file with intensities to be normalised.
    config : inspire.config.Config
        Config object for the whole experiment.

    Returns
    -------
    quant_df : pd.DataFrame
        DataFrame of quantifications, now including normalised intensities.
    """
    # Split between valid and invalid quantifications:
    valid_quant_df = quant_df[quant_df[f'{source}_clean'].notna()]
    invalid_quant_df = quant_df[quant_df[f'{source}_clean'].isna()]

    valid_quant_df[f'{source}_norm'] = (
        valid_quant_df[f'{source}_clean'] - valid_quant_df[f'{source}_clean'].median()
    ) + global_median

    # Fill with null for invalid samples and combine all entries
    invalid_quant_df[f'{source}_norm'] = np.nan
    quant_df = pd.concat([valid_quant_df, invalid_quant_df])

    return quant_df



""" Functions for reading in MSFragger search results.
"""
import multiprocessing
import os
from pathlib import Path

import polars as pl
import pandas as pd
from pyteomics import pepxml

from inspire.constants import (
    ACCESSION_KEY,
    CHARGE_KEY,
    DELTA_SCORE_KEY,
    ENGINE_SCORE_KEY,
    KNOWN_PTM_WEIGHTS,
    LABEL_KEY,
    MASS_DIFF_KEY,
    PEPTIDE_KEY,
    PTM_ID_KEY,
    PTM_IS_VAR_KEY,
    PTM_NAME_KEY,
    PTM_SEQ_KEY,
    PTM_WEIGHT_KEY,
    RT_KEY,
    SCAN_KEY,
    SEQ_LEN_KEY,
    SOURCE_KEY,
)
from inspire.utils import filter_for_prosit

# Define the relevant column names from MSFragger search results.
MSF_ACCESSION_KEY = 'Proteins'
MSF_DELTA_KEY = 'delta_hyperscore'
MSF_PPM_ERR_KEY = 'abs_ppm'
MSF_PEPTIDE_KEY = 'Peptide'
MSF_MASS_KEY = 'ExpMass'
MSF_LABEL_KEY = 'Label'
MSF_PEP_LEN_KEY = 'peptide_length'
MSF_SCORE_KEY = 'hyperscore'
MSF_RT_KEY = 'retentiontime'
MSF_LOG10E_KEY = 'log10_evalue'
MSF_NTT_KEY = 'ntt'
MSF_NMC_KEY = 'nmc'

MSF_IGNORE_COLS = [
    MSF_LOG10E_KEY,
    MSF_NMC_KEY,
    MSF_NTT_KEY,
]

ID_NUMBERS = {
    'unknown': 6,
    'Deamidation (N)': 5,
    'Deamidation (Q)': 5,
    'Phospho (S)': 4,
    'Phospho (T)': 4,
    'Phospho (Y)': 4,
    'Acetyl (N-term)': 3,
    'Oxidation (M)': 2,
    'Carbamidomethyl (C)': 1,
}

MSF_MAPPING_PEP_XML = {
    '115': 'Deamidation (N)',
    '129': 'Deamidation (Q)',
    '167': 'Phospho (S)',
    '181': 'Phospho (T)',
    '243': 'Phospho (Y)',
    '43': 'Acetyl (N-term)',
    '147': 'Oxidation (M)',
    '160': 'Carbamidomethyl (C)',
}

MSF_MAPPING_PEP_XML_REV = {
    item: key for key, item in MSF_MAPPING_PEP_XML.items()
}

def _extract_psm_id_data(spec_id):
    """ Function to extract data from the PSM ID column to standard inSPIRE format.

    Parameters
    ----------
    spec_id : str
        The spectrum ID from MSFragger pepXML output.

    Returns
    -------
    results : dict
        Dictionary of the scan, source and reported charge of the peptide.
    """
    results = {}
    spec_id_data = spec_id.split('.')
    results[SOURCE_KEY] = spec_id_data[0]
    results[SCAN_KEY] = int(spec_id_data[1])
    results[f'alt_{CHARGE_KEY}'] = int(spec_id_data[-1].split('_')[0])
    return results

def _extract_mod_weight(code):
    """ Function to extract the weight of a modification.

    Parameters
    ----------
    code : dict
        Dictionary describing a modification.

    Returns
    -------
    mass : int or None
    """
    if code['mass'] is not None:
        return round(code['mass'])
    return None


def _get_msf_mods_pepxml(msf_df, fixed_modifications):
    """ Function to get the modifications based on the MS fragger pepXML columns.

    Parameters
    ----------
    msf_df : pd.DataFrame
        MSFragger results read in from pepXML format.
    fixed_modifications : list of str or None
        A list of the fixed modification used in the exeriment.

    Returns
    -------
    mod_df : pd.DataFrame
        Small DataFrame of the modifications found in the data.
    msf_name_to_id : dict
        A dictionary mapping names of modifications in MSFragger to inSPIRE IDs.
    """
    if 'modifications' not in msf_df.columns:
        return pd.DataFrame(), {}

    msf_mods = sorted(
        msf_df['modifications'].explode().apply(
            _extract_mod_weight,
            skip_nulls=False,
        ).drop_nulls().unique().to_list()
    )
    msf_mods = [str(int(x)) for x in msf_mods if int(x) != -1]

    mod_names = [MSF_MAPPING_PEP_XML.get(msf_mod, 'unknown') for msf_mod in msf_mods]
    mod_weights = [KNOWN_PTM_WEIGHTS.get(mod, 100) for mod in mod_names]

    mod_df = pd.DataFrame({
        PTM_NAME_KEY: mod_names,
        PTM_WEIGHT_KEY: mod_weights,
        PTM_IS_VAR_KEY: [
            fixed_modifications is None or not nm in fixed_modifications for nm in mod_names
        ],
        'msfMod': msf_mods,
    })

    mod_df = mod_df.sort_values(by=PTM_NAME_KEY)
    mod_df = mod_df.reset_index(drop=True)
    mod_df[PTM_ID_KEY] = mod_df[PTM_NAME_KEY].apply(ID_NUMBERS.get)
    msf_name_to_id = dict(zip(
        mod_df['msfMod'].tolist(),
        [str(x) for x in mod_df[PTM_ID_KEY].tolist()],
    ))
    return mod_df, msf_name_to_id

def _separate_msf_ptms(mod_pep, ms_frag_mappings):
    """ Function to separate MSFragger reported PTMs.

    Parameters
    ----------
    df_row : pd.Series
        Row of MSFragger results DataFrame
    ms_frag_mappings : dict
        Dictionary mapping PTM weights to their inSPIRE ID.

    Returns
    -------
    df_row : pd.Series
        The input row with peptide sequence and modifications separated.
    """
    peptide = mod_pep
    start_mod = '0'
    if peptide[0:5] == 'n[43]':
        start_mod = '3'
        peptide = peptide[5:]
    if '[' in peptide:
        pep_seq = ''
        ptm_seq = ''
        while peptide:
            assert peptide[0] != '['
            if len(peptide) == 1:
                pep_seq += peptide[0]
                ptm_seq += '0'
                break

            if peptide[1] != '[':
                pep_seq += peptide[0]
                ptm_seq += '0'
                peptide = peptide[1:]
            else:
                pep_seq += peptide[0]
                peptide = peptide[1:]

                end_mod_str = peptide.index(']') + 1
                ptm_seq += ms_frag_mappings[peptide[1:end_mod_str-1]]
                peptide = peptide[end_mod_str:]

        return f'{start_mod}.{ptm_seq}.0'
    if start_mod == '3':
        return f'{start_mod}.{"0"*len(peptide)}.0'
    return None

def flatten_protein_data(proteins):
    """ Function to extract protein accession data from pepXML dictionary.

    Parameters
    ----------
    proteins : pd.Series
        The pepXML protein data for a single hit.

    Returns
    -------
    df_row : pd.Series
        The input data updated to parse protein accession data.
    """
    label = -1
    results = {}
    new_proteins = []
    for protein in proteins:
        if not protein['protein'].startswith('rev'):
            label = 1
        new_proteins.append(protein['protein'])
    results[LABEL_KEY] = label
    results[ACCESSION_KEY] = ','.join(new_proteins)
    return results

def get_read_path(df_loc, file_idx):
    """ Function to remove invalid characters from a source.

    Parameters
    ----------
    df_loc : str
        Location of the file.
    file_idx : int
        The index of the input file.

    Returns
    -------
    read_loc : str
        The final location that should be read from.
    remove_read_loc : bool
        Flag indicating if a temporary file was generated which should be removed.
    """
    source_name = Path(df_loc).name
    if source_name.endswith('_uncalibrated.pepXML'):
        source_name = source_name[:-20]
    elif source_name.endswith('_calibrated.pepXML'):
        source_name = source_name[:-18]
    elif source_name.endswith('.pepXML'):
        source_name = source_name[:-7]
    if "'" in source_name:
        with open(df_loc, 'r', encoding='UTF-8') as file :
            filedata = file.read()
        filedata = filedata.replace(source_name, f'temp_{file_idx}')
        with open(df_loc.replace(source_name, f'temp_{file_idx}'), 'w', encoding='UTF-8') as file:
            file.write(filedata)
        return df_loc.replace(source_name, f'temp_{file_idx}'), True
    return df_loc, False

def read_single_ms_fragger_data(df_loc, fixed_modifications, file_idx):
    """ Function to read in PEAKS DB search results from a single file.

    Parameters
    ----------
    df_loc : str
        A location of PEAKS DB search results.

    Returns
    -------
    hits_df : pd.DataFrame
        A DataFrame of all search results properly formatted for inSPIRE.
    mods_dfs : pd.DataFrame
        A small DataFrame detailing the ptms found in the data.
    """
    read_loc, remove_read_loc  = get_read_path(df_loc, file_idx)

    all_psms = []
    with pepxml.read(read_loc) as psms:
        for psm in psms:
            psms_to_add = []
            peps_added = []

            for pep_psm in psm['search_hit']:
                if pep_psm['peptide'] not in peps_added:
                    pep_psm[CHARGE_KEY] = int(psm['assumed_charge'])
                    pep_psm['SpecId'] = psm['spectrum']
                    pep_psm[MSF_RT_KEY] = psm['retention_time_sec']
                    pep_psm['hyperscore'] = pep_psm['search_score']['hyperscore']
                    del pep_psm['search_score']
                    peps_added.append(pep_psm['peptide'])
                    psms_to_add.append(pep_psm)

            if len(psms_to_add) > 1:
                rank_2_score = psms_to_add[1]['hyperscore']
            else:
                rank_2_score = 0

            for pep_psm in psms_to_add:
                pep_psm[MSF_DELTA_KEY] = pep_psm['hyperscore'] - rank_2_score
                all_psms.append(pep_psm)

    msf_df = pl.DataFrame(all_psms)

    msf_df = msf_df.with_columns(
        pl.col('SpecId').apply(
            _extract_psm_id_data
        ).alias('results')
    ).unnest('results')

    if CHARGE_KEY not in msf_df.columns:
        msf_df = msf_df.rename({f'alt_{CHARGE_KEY}': CHARGE_KEY})

    # Separate PTMs.
    var_mod_df, msf_name_to_id = _get_msf_mods_pepxml(msf_df, fixed_modifications)

    msf_df = msf_df.with_columns(
        pl.col('modified_peptide').apply(
            lambda mod_pep : _separate_msf_ptms(mod_pep, msf_name_to_id)
        ).alias(PTM_SEQ_KEY)
    )

    msf_df = msf_df.with_columns(
        pl.col('proteins').apply(flatten_protein_data).alias('results')
    )
    msf_df = msf_df.drop('proteins')
    msf_df = msf_df.unnest('results')

    msf_df= msf_df.with_columns(
        pl.col(PEPTIDE_KEY).apply(len).alias(MSF_PEP_LEN_KEY),
    )
    msf_df= msf_df.with_columns(
        (
            pl.col('calc_neutral_pep_mass')/pl.col(MSF_PEP_LEN_KEY)
        ).alias('avgResidueMass'),
    )

    # Rename to match inSPIRE naming scheme.
    msf_df = msf_df.rename({
        MSF_PEP_LEN_KEY: SEQ_LEN_KEY,
        'massdiff': MASS_DIFF_KEY,
        MSF_RT_KEY: RT_KEY,
        MSF_SCORE_KEY: ENGINE_SCORE_KEY,
        MSF_DELTA_KEY: DELTA_SCORE_KEY,
    })

    msf_df = msf_df.with_columns(
        pl.lit(0).alias('fromChimera'),
        pl.lit(0).alias('missedCleavages'),
        pl.col(MASS_DIFF_KEY).apply(abs).alias(MASS_DIFF_KEY),
    )

    # Filter for Prosit and add feature columns not present.
    msf_df = filter_for_prosit(msf_df)

    msf_df = msf_df.select(
        'source',
        'scan',
        'peptide',
        'Label',
        'proteins',
        'ptm_seq',
        'sequenceLength',
        'missedCleavages',
        'charge',
        'massDiff',
        'retentionTime',
        'engineScore',
        'deltaScore',
        'fromChimera',
        'avgResidueMass',
    )

    if remove_read_loc:
        os.remove(read_loc)

    return msf_df, var_mod_df

def read_ms_fragger_data(ms_fragger_data, fixed_modifications, n_cores, reduce_results):
    """ Function to read in PEAKS DB search results from one or more files.

    Parameters
    ----------
    peaks_data : str or list of str
        A single location of PEAKS DB search results or a list of locations.

    Returns
    -------
    hits_df : pd.DataFrame
        A DataFrame of all search results properly formatted for inSPIRE.
    mods_dfs : pd.DataFrame
        A small DataFrame detailing the ptms found in the data.
    """
    if isinstance(ms_fragger_data, list):
        func_args = [
            (msf_file, fixed_modifications, file_idx) for file_idx, msf_file in enumerate(
                ms_fragger_data
            )
        ]

        with multiprocessing.get_context('spawn').Pool(processes=n_cores) as pool:
            results = pool.starmap(read_single_ms_fragger_data, func_args)

        mods_dfs = [res[1] for res in results]

        # Combine DataFrames and validate that same PTMs are present.
        hits_df = pl.concat([res[0] for res in results])

        if reduce_results:
            hits_df = hits_df.sort(by=[ENGINE_SCORE_KEY, LABEL_KEY], descending=True)
            hits_df = hits_df.unique(subset=[SOURCE_KEY, SCAN_KEY])

        return hits_df, mods_dfs[0]

    hits_df, mods_df = read_single_ms_fragger_data(ms_fragger_data, fixed_modifications, 0)
    if reduce_results:
        hits_df = hits_df.sort(by=[ENGINE_SCORE_KEY, LABEL_KEY], descending=True)
        hits_df = hits_df.unique(subset=[SOURCE_KEY, SCAN_KEY])

    return hits_df, mods_df



""" Functions for generating the html report at the end of the inSPIRE
    pipeline.
"""
import os
import webbrowser

from inspire.constants import ENDC_TEXT, OKCYAN_TEXT

def create_html_report(config, figures):
    """ Function to create the final html report and open it in the brower.

    Parameters
    ----------
    config : inspire.config.Config
        The Config object for the whole pipeline.
    figures : dict
        A dictionary containing all of the required plots.
    """
    config_str = str(config)
    html_string = ('''
    <html>
        <head>
            <link 
                rel="stylesheet"
                href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.1/css/bootstrap.min.css"
            >
            <style>
                body{
                    margin:0 100;
                    padding-bottom: 50px;
                    background:whitesmoke;
                }
                h2{
                    color: firebrick;
                }
                h3{
                    color: firebrick;
                }
                h2:after
                {
                    content:' ';
                    display: block;
                    border:0.25px solid #696969;
                    position: absolute;
                    width: 60%;
                    margin-top: 2px;
                    left: 20%;
                }
                table {
                    width: 60%;
                    border: 2px solid #696969;
                }
                th, td {
                    border: 1px solid #696969;
                    padding: 2px;
                }
            </style>
        </head>
        <body>
            <center>
            <h2>inSPIRE Report for ''' + config.experiment_title + '''</h2>
            </center>
            <h3>
                inSPIRE Settings Used
            </h3>
            <p>
                ''' + config_str + '''
            </p>
            <h3>
                Selected Features and Importance
            </h3>
            <p>
                The table below shows the importance of the final feature set used by
                percolator. Strong positive values (highlighted in green) may indicate that
                higher feature values are more common among target PSMs while strongly negative
                values (highlighted in red) may indicate that higher feature values are more
                common among decoy PSMs. However, it is also possible that a feature like
                searchEngineScore ends up with a negative coefficient simply because it
                is so strongly correlated to a more powerful feature like deltaScore which
                has a strong positive coefficent.
            </p>
            <center>
    ''' + figures['table'] +
    '''
            </center>
            <h3>
                Feature Distributions
            </h3>
            <p>
                These Violin Plots show the distributions of the three most heavily positive
                and heavily negative weighted features for accepted and rejected PSMs.
            </p>
            <center>
    ''' + figures['violin_fig'])

    html_string += (
        '''
                </center>
                <h3>
                    inSPIRE Performance: Number of PSMs Identified
                </h3>
                <p>
                    This shows the number of PSMs discovered by inSPIRE compared to the original
                    search engine for q-value cut offs between 0.01 and 0.1.
                </p>
                <center>
        ''' + figures['psms_fig']
    )

    if config.use_binding_affinity in ['asValidation', 'asFeature']:
        html_string += ('''
                </center>
                <h3>
                    inSPIRE Performance: Percentage Binders Identified
                </h3>
                <p>
                    This shows the percentage of HLA-I binders as predicted by NetMHCpan among
                    the PSMs identified by inSPIRE compared to the original search engine
                    results. Note that this plot is more meaningful if you have set
                    useBindingAffinty to asValidation, as if it is set to asFeature it is
                    unsurprising that inSPIRE would produce a higher percentage of HLA binders.
                </p>
                <center>
        ''' + figures['binders_fig']
    )

    html_string += ('''
            </center>
        </body>
    </html>
    ''')

    output_path = f'{config.output_folder}/inspire-report.html'
    with open(output_path, 'w', encoding='UTF-8') as output_file:
        output_file.write(html_string)

    print(
        OKCYAN_TEXT +
        '\tReport generated.' +
        ENDC_TEXT
    )

    if not config.silent_execution:
        webbrowser.open(
            'file://' + os.path.realpath(output_path),
            new=2
        )



""" Functions for calibrating the optimal collision energy setting
    for your experiment.
"""
import polars as pl

from inspire.constants import (
    ACCESSION_STRATUM_KEY,
    BOLD_TEXT,
    ENDC_TEXT,
    ENGINE_SCORE_KEY,
    KNOWN_PTM_WEIGHTS,
    LABEL_KEY,
    OKBLUE_TEXT,
    OKCYAN_TEXT,
    PTM_ID_KEY,
    PTM_NAME_KEY,
    PTM_SEQ_KEY,
    SCAN_KEY,
    SEQ_LEN_KEY,
    SOURCE_KEY,
    SPECTRAL_ANGLE_KEY,
    UNDERLINE_TEXT,
    CHARGE_KEY,
    INTENSITIES_KEY,
    MZS_KEY,
    PEPTIDE_KEY,
    PROSIT_INTES_KEY,
    PROSIT_IONS_KEY,
    PROSIT_SEQ_KEY,
)
from inspire.input.mgf import process_mgf_file
from inspire.input.msp import msp_to_df
from inspire.input.mzml import process_mzml_file
from inspire.input.search_results import generic_read_df
from inspire.feature_creation import combine_spectral_data
from inspire.predict_spectra import predict_spectra
from inspire.prepare import write_prosit_input_df
from inspire.spectral_features import calculate_spectral_features
from inspire.utils import (
    check_bad_mods,
    get_ox_flag,
    get_cam_flag,
    remove_source_suffixes,
)

COLLISION_ENERGY_RANGE = [20 + i for i in range(21)]

def _get_top_hits(config):
    """ Function to extract the top scoring hits for collision energy calibration.

    Parameters
    ----------
    config : inspire.config.Config
        The Config object for the experiment.
    """
    target_df, mods_df = generic_read_df(config, save_dfs=False, for_calibration=True)

    unknown_mods = mods_df[
        (mods_df[PTM_NAME_KEY] != 'Oxidation (M)') &
        ((mods_df[PTM_NAME_KEY] != 'Carbamidomethyl (C)'))
    ][PTM_ID_KEY].tolist()
    unknown_mods = {str(x) for x in unknown_mods}

    target_df = target_df.with_columns(
        pl.col(PTM_SEQ_KEY).apply(
            lambda x : check_bad_mods(x, unknown_mods),
            skip_nulls=False,
        ).alias('unknownModifications')
    )
    target_df = target_df.filter(
        ~pl.col('unknownModifications')
    ).drop(
        ['unknownModifications'],
    )

    if ACCESSION_STRATUM_KEY in target_df.columns:
        target_df = target_df.filter(pl.col(ACCESSION_STRATUM_KEY).eq(0))

    target_df = target_df.filter(
        target_df[LABEL_KEY] == 1
    )

    if config.rescore_method == 'percolatorSeparate':
        score_cut_off = target_df[ENGINE_SCORE_KEY].quantile(0.98)
    else:
        score_cut_off = target_df[ENGINE_SCORE_KEY].quantile(0.90)

    target_df = target_df.filter(
        (target_df[ENGINE_SCORE_KEY] > score_cut_off)
    )

    if target_df.shape[0] > 600:
        target_df = target_df.top_k(
            600,
            by=[ENGINE_SCORE_KEY, SEQ_LEN_KEY, SCAN_KEY, PEPTIDE_KEY],
            descending=True,
        )

    return target_df, mods_df

def prepare_calibration(config):
    """ Function to generate Prosit input for collision energy calibration.

    Parameters
    ----------
    config : inspire.config.Config
        The Config object for the experiment.
    """
    target_df, mods_df = _get_top_hits(config)

    for idx, collision_energy in enumerate(COLLISION_ENERGY_RANGE):
        write_prosit_input_df(
            target_df,
            mods_df,
            config,
            collision_energy,
            'calibrationInput',
            overwrite=idx==0,
        )

    return target_df, mods_df

def calibrate(config):
    """ Function to calibrate the optimal collision energy for Prosit input.

    Parameters
    ----------
    config : inspire.config.Config
        The Config object for the experiment.
    """
    print(
        OKCYAN_TEXT +
        '\tSelecting top hits...' +
        ENDC_TEXT
    )
    target_df, mods_df = prepare_calibration(config)
    predict_spectra(config, 'calibrate')

    prosit_df = msp_to_df(
        f'{config.output_folder}/calibrationPredictions.msp', 'prosit', None
    )

    scan_files = sorted(target_df[SOURCE_KEY].unique().to_list())

    ox_flag = get_ox_flag(mods_df)
    cam_flag = get_cam_flag(mods_df)

    mods_dict = {
        0: 0.0,
        int(cam_flag): KNOWN_PTM_WEIGHTS['Carbamidomethyl (C)'],
        int(ox_flag): KNOWN_PTM_WEIGHTS['Oxidation (M)']
    }

    scan_dfs = []
    for scan_file in scan_files:
        filtered_search_df = target_df.filter(
            target_df[SOURCE_KEY].eq(scan_file)
        )

        scans = filtered_search_df[SCAN_KEY].unique()
        if config.scans_format == 'mzML':
            scan_df = process_mzml_file(
                f'{config.scans_folder}/{scan_file}.{config.scans_format}',
                set(scans.to_list()),
            )
        else:
            scan_df = process_mgf_file(
                f'{config.scans_folder}/{scan_file}.{config.scans_format}',
                set(scans.to_list()),
                config.scan_title_format,
                config.source_files,
                combined_source_file=config.combined_scans_file is not None,
            )
        scan_dfs.append(scan_df.unique(subset=[SOURCE_KEY, SCAN_KEY]))
    combined_scan_df = pl.concat(scan_dfs)
    print(
        OKCYAN_TEXT +
        '\t\tCombining all spectral data...' +
        ENDC_TEXT
    )
    combined_df = combine_spectral_data(
        target_df,
        combined_scan_df,
        prosit_df,
        ox_flag,
        'prosit',
    )
    print(
        OKCYAN_TEXT +
        '\t\tCalculating spectral angles...' +
        ENDC_TEXT
    )
    combined_df = combined_df.with_columns(
         pl.struct([
            CHARGE_KEY,
            'collisionEnergy',
            INTENSITIES_KEY,
            MZS_KEY,
            PEPTIDE_KEY,
            PROSIT_INTES_KEY,
            PROSIT_IONS_KEY,
            PROSIT_SEQ_KEY,
            PTM_SEQ_KEY,
        ]).apply(
            lambda x : calculate_spectral_features(
                x,
                mods_dict,
                config.mz_accuracy,
                config.mz_units,
                None,
                '1',
                config.delta_method,
                minimal_features=True,
            ),
            skip_nulls=False,
        ).alias('spectralResults')
    ).unnest('spectralResults')

    results_df = combined_df.groupby('collisionEnergy', maintain_order=True).agg(
        pl.mean(SPECTRAL_ANGLE_KEY)
    )

    optimal_collision_energy = results_df.sort(
        by='spectralAngle', descending=True,
    )['collisionEnergy'].to_list()[0]

    print(
        OKBLUE_TEXT + BOLD_TEXT + UNDERLINE_TEXT +
        f'\n\t---> Optimal Collision Energy Setting: {optimal_collision_energy} <---\n' +
        ENDC_TEXT
    )
    results_df = results_df.sort(by='collisionEnergy')
    results_df.write_csv(
        f'{config.output_folder}/collisionEnergyStats.csv'
    )



""" Functions for reading in Prosit predicted spectra in msp format.
"""
import re

import numpy as np
import polars as pl

from inspire.constants import (
    CHARGE_KEY,
    MS2PIP_NAME_MAPPINGS,
    OXIDATION_PREFIX,
    OXIDATION_PREFIX_LEN,
    PROSIT_IONS_KEY,
    PROSIT_INTES_KEY,
    PROSIT_SEQ_KEY,
    PTM_ID_KEY,
    PTM_NAME_KEY,
    PTM_SEQ_KEY,
)

def msp_process_sequence_and_charge(line):
    """ Function to extract the name and charge of a sample from
        the relevant line of an msp file.

    Parameters
    ----------
    line : str
        The latest line read from the msp file.

    Returns
    -------
    sequence : str
        The peptide sequence.
    charge : int
        The charge of the sequence.
    """
    regex_match = re.match(r'Name: (.*?)/(\d+)\n', line)
    sequence = regex_match.group(1)
    charge = int(regex_match.group(2))
    return sequence, charge

def process_intensities(intensities):
    """ Function to L2 normalise intensity array.

    Parameters
    ----------
    intensities : list of float
        A list of the intensities of peaks found in the spectrum.

    Returns
    -------
    normed_intensities : list of float
        A list of normalised intensities.
    """
    l2_norm = np.linalg.norm(np.array(intensities), ord=2)
    normed_intensities = [z/l2_norm for z in intensities]
    return normed_intensities

def msp_process_peaks(line, msp_file):
    """ Function to extract the ms2 spectrum of a sample from
        the relevant lines of an msp file.

    Parameters
    ----------
    line : str
        The latest line read from the msp file.
    msp_file : file
        The full msp file.

    Returns
    -------
    ions : list of str
        A list of the ions predicted
    normed_intensities : list of float
        A list of ions normed intensities.
    """
    regex_match = re.match(r'Num peaks: (\d+)\n', line)
    n_peaks = int(regex_match.group(1))

    ions = []
    intensities = []
    for _ in range(n_peaks):
        line = msp_file.readline()
        peak_data = line.strip().split('\t')
        intensity = float(peak_data[1])
        regex_match = re.match(r'(.*?)/(.*?)ppm', peak_data[2].strip('"'))
        ion = regex_match.group(1).strip(')').strip('(')
        if intensity > 0:
            ions.append(ion)
            intensities.append(intensity)

    normed_intensities = process_intensities(intensities)

    return ions, normed_intensities

def ms2pip_process_peaks(line, msp_file):
    """ Function to extract the ms2 spectrum of a sample from
        the relevant lines of an msp file.

    Parameters
    ----------
    line : str
        The latest line read from the msp file.
    msp_file : file
        The full msp file.

    Returns
    -------
    normed_intensities : dict
        A dictionary of ion names mapped to their normed intensity.
    """
    regex_match = re.match(r'Num peaks: (\d+)\n', line)
    n_peaks = int(regex_match.group(1)) - 1

    ions = []
    intensities = []
    for _ in range(n_peaks):
        line = msp_file.readline()

        peak_data = line.strip().split('\t')

        if len(peak_data) == 1:
            break
        intensity = float(peak_data[1])
        ion = peak_data[2].strip('"')
        if intensity > 0:
            ions.append(ion)
            intensities.append(intensity)

    normed_intensities = process_intensities(intensities)

    return ions, normed_intensities

def get_ms2pip_mods(line, sequence, mod_id_mappings):
    """ Function to fetch the modifications from an MS2PIP modification file.

    Parameters
    ----------
    line : str
        A line from the msp file.
    sequence : str
        The peptide sequence.
    mod_id_mappings : dict
        Dictionary mapping modification names to their single digit code.

    Returns
    -------
    ptm_seq : str
        The PTM sequence that encodes all modifications in the data.
    """
    regex_match = re.match(
        r'Comment: Mods=(.*?) Parent=(.*?)',
        line,
    )
    modifications = regex_match.group(1)
    mod_list = modifications.split('/')[1:]

    if not mod_list:
        return None
    ptm_seq = '0'*(len(sequence)+2)

    for entry in mod_list:
        ptm_data = entry.split(',')
        ptm_name = ptm_data[2]
        ptm_loc = int(ptm_data[0])
        ptm_loc += 1
        ptm_seq = ptm_seq[:ptm_loc] + str(mod_id_mappings[ptm_name]) + ptm_seq[ptm_loc+1:]

    ptm_seq = ptm_seq[0] + '.' + ptm_seq[1:-1] + '.' + ptm_seq[-1]

    return ptm_seq

def process_prosit_comment(line, sequence):
    """ Function to extract data the comments on a sample from
        the relevant line of an msp file of Prosit output.

    Parameters
    ----------
    line : str
        The latest comment line read from the msp file.
    sequence : str
        The latest peptide sequence read from the msp file.

    Returns
    -------
    sequence : str
        The sequence with any modifications added.
    irt : float
        The iRT value predicted by Prosit.
    collision_energy : int
        The collision energy setting used by Prosit.
    """
    regex_match = re.match(
        r'(.*?) Collision_energy=(.*?) Mods=(.*?) ModString=(.*?)//(.*?)/(.*?)',
        line
    )
    collision_energy = int(float(regex_match.group(2)))
    mods = regex_match.group(5)
    if 'iRT=' in line:
        irt = float(line.split('iRT=')[-1].split(' ')[0])
    else:
        irt = None

    if mods:
        mods_list = mods.split('; ')
        mod_seq = ''
        previous_ind = 0
        for mod in mods_list:
            if mod.startswith(OXIDATION_PREFIX):
                pos = int(mod[OXIDATION_PREFIX_LEN:]) - 1
                mod_seq += sequence[previous_ind:pos]
                mod_seq += "(ox)"
                previous_ind = pos
        mod_seq += sequence[previous_ind:]
        return mod_seq, irt, collision_energy

    return sequence, irt, collision_energy

def msp_to_df(msp_filename, msp_format, mods_df):
    """ Function to process an msp file and extract relevant information
        for training into csv format (tab separated).

    Parameters
    ----------
    msp_filename : str
        The location where the msp file is written.
    msp_format : str
        Either prosit or ms2pip. The predictor which wrote the msp file.
    mods_df : pd.DataFrame
        Small DataFrame detailing the PTMs considered.

    Returns
    -------
    ion_df : pd.DataFrame
        The DataFrame with the spectra found in the msp file.
    """
    if msp_format == 'ms2pip':
        mods_df['ms2pipName'] = mods_df[PTM_NAME_KEY].apply(
            lambda x : MS2PIP_NAME_MAPPINGS[x]
        )
        mod_id_mappings = dict(zip(mods_df['ms2pipName'].tolist(), mods_df[PTM_ID_KEY].tolist(), ))

    with open(msp_filename, 'r', encoding='UTF-8') as msp_file:
        peptides = []
        charges = []
        ion_intensities = []
        modified_sequences = []
        irts = []
        collision_energies = []
        ptm_seqs = []
        ion_names = []

        while (line := msp_file.readline()):
            if line.startswith('Name: '):
                sequence, charge = msp_process_sequence_and_charge(line)

                line = msp_file.readline()
                assert line.startswith('MW: ')

                line = msp_file.readline()
                assert line.startswith('Comment: ')

                if msp_format == 'prosit':
                    modified_sequence, irt, col_e = process_prosit_comment(line, sequence)
                else:
                    modified_sequence = sequence
                    ptm_seq = get_ms2pip_mods(line, sequence, mod_id_mappings)
                line = msp_file.readline()
                assert line.startswith('Num peaks: ')

                if msp_format == 'prosit':
                    named_ions, normed_intensities = msp_process_peaks(line, msp_file)
                else:
                    named_ions, normed_intensities = ms2pip_process_peaks(line, msp_file)

                ion_intensities.append(normed_intensities)
                ion_names.append(named_ions)
                peptides.append(sequence)
                charges.append(charge)
                modified_sequences.append(modified_sequence)
                if msp_format == 'prosit':
                    irts.append(irt)
                    collision_energies.append(col_e)
                else:
                    ptm_seqs.append(ptm_seq)

    df_data = {
        CHARGE_KEY: charges,
        PROSIT_IONS_KEY: ion_names,
        PROSIT_INTES_KEY: ion_intensities,
    }
    if msp_format == 'prosit':
        df_data[PROSIT_SEQ_KEY] = modified_sequences
        df_data['iRT'] = irts
        df_data['collisionEnergy'] = collision_energies
    else:
        df_data['peptide'] = modified_sequences
        df_data[PTM_SEQ_KEY] = ptm_seqs

    msp_df = pl.DataFrame(df_data)

    return msp_df
